import Appier

public class AppierFlutter {
  public static let shared = AppierFlutter()
  
  public func configure(_ arguments: [String: Any], result: FlutterResult) {
    guard let appId = arguments["appId"] as? String else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "configure: appId should not be null",
        details: nil
      ))
      return
    }

    let appGroup = (arguments["appGroup"] as? String) ?? ""
    let isDev = (arguments["isDev"] as? Bool) ?? false
    guard let version = arguments["version"] as? String else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "configure: version should not be null",
        details: nil
      ))
      return
    }
    QGSdk.getSharedInstance().onStart(
      appId,
      withAppGroup: appGroup,
      setDevProfile: isDev,
      frameworkType: "flutter",
      version: version
    )
    result(nil)
  }
  
  public func renewAppierId(_ arguments: [String: Any], result: @escaping FlutterResult) {
    QGSdk.getSharedInstance().renewAppierId({
      result(nil)
    })
  }

  public func getAppierId(_ arguments: [String: Any], result: FlutterResult) {
    result(QGSdk.getSharedInstance().getAppierId())
  }

  public func setPushToken(_ arguments: [String: Any], result: FlutterResult) {
    guard let token = arguments["token"] as? String else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): token should not be null",
        details: nil
      ))
      return
    }
    // This is not a bug or a typo to connect `setPushToken` to `setFCMToken`. Actually we are connecting it to the `setGCMId` but it's not exposed.
    QGSdk.getSharedInstance().setFCMToken(token)
    result(nil)
  }

  public func setIDFAConsent(_ arguments: [String: Any], result: FlutterResult) {
    guard let idfaConsent = arguments["idfaConsent"] as? Bool else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): idfaConsent should not be null",
        details: nil
      ))
      return
    }
    QGSdk.getSharedInstance().setIDFAConsent(idfaConsent)
    result(nil)
  }

  public func logEvent(_ arguments: [String: Any], result: FlutterResult) {
    guard let eventName = arguments["eventName"] as? String else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "logEvent: eventName should not be null",
        details: nil
      ))
      return
    }
    
    let parameters = (arguments["parameters"] as? [String: Any]) ?? [: ]
    let vts = arguments["vts"] as? NSNumber
    let vtsCurr = arguments["vtsCurr"] as? String
    QGSdk.getSharedInstance().logEvent(
      eventName,
      withParameters: parameters,
      withValueToSum: vts,
      withValueToSumCurrency: vtsCurr
    )
    result(nil)
  }
  
  public func setUniversalLinkDomains(_ arguments: [String: Any], result: FlutterResult) {
    guard let linkDomains = arguments["linkDomains"] as? [String], !linkDomains.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setUniversalLinkDomains: linkDomains should not be null or empty",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setUniversalLinkDomains(linkDomains)
    result(nil)
  }
  
  public func flush(_ arguments: [String: Any], result: FlutterResult) {
    QGSdk.getSharedInstance().flush()
    result(nil)
  }
  
  public func setUserId(_ arguments: [String: Any], result: FlutterResult) {
    guard let userId = arguments["userId"] as? String, !userId.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setUserId: userId should not be null or empty",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setUserId(userId)
    result(nil)
  }
  
  public func setName(_ arguments: [String: Any], result: FlutterResult) {
    guard let name = arguments["name"] as? String, !name.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setName: name should not be null or empty",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setName(name)
    result(nil)
  }
  
  public func setFirstName(_ arguments: [String: Any], result: FlutterResult) {
    guard let firstName = arguments["firstName"] as? String, !firstName.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setFirstName: firstName should not be null or empty",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setFirstName(firstName)
    result(nil)
  }
  
  public func setLastName(_ arguments: [String: Any], result: FlutterResult) {
    guard let lastName = arguments["lastName"] as? String, !lastName.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setLastName: lastName should not be null or empty",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setLastName(lastName)
    result(nil)
  }
  
  public func setCity(_ arguments: [String: Any], result: FlutterResult) {
    guard let city = arguments["city"] as? String, !city.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setCity: city should not be null or empty",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setCity(city)
    result(nil)
  }
  
  public func setEmail(_ arguments: [String: Any], result: FlutterResult) {
    guard let email = arguments["email"] as? String, !email.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setEmail: email should not be null or empty",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setEmail(email)
    result(nil)
  }
  
  public func setPhoneNumber(_ arguments: [String: Any], result: FlutterResult) {
    guard let phoneNo = arguments["phoneNo"] as? String, !phoneNo.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setPhoneNumber: phoneNo should not be null or empty",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setCustomKey("phoneNo", withValue:phoneNo)
    result(nil)
  }
  
  public func setDayOfBirth(_ arguments: [String: Any], result: FlutterResult) {
    guard let day = arguments["day"] as? NSNumber, day.intValue >= 1 else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setDayOfBirth: day should not be null or non-positive",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setDayOfBirth(day)
    result(nil)
  }
  
  public func setMonthOfBirth(_ arguments: [String: Any], result: FlutterResult) {
    guard let month = arguments["month"] as? NSNumber, month.intValue >= 1 else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setMonthOfBirth: month should not be null or non-positive",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setMonthOfBirth(month)
    result(nil)
  }
  
  public func setYearOfBirth(_ arguments: [String: Any], result: FlutterResult) {
    guard let year = arguments["year"] as? NSNumber, year.intValue >= 1 else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setYearOfBirth: year should not be null or non-positive",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setYearOfBirth(year)
    result(nil)
  }
  
  public func setCustomKey(_ arguments: [String: Any], result: FlutterResult) {
    // FIXME:
    // Currently null value is not accepted as the setCustomKey(_:withValue:) of the Swift API
    // even though it should be because that's the only way to delete value for the key.
    // Fix the logic to accept null value when the Swift API starts accepting it.
    guard let key = arguments["key"] as? String, !key.isEmpty, let value = arguments["value"] else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setCustomKey: key should not be null or empty",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setCustomKey(key, withValue: value)
    result(nil)
  }
  
  public func setAttributionWindow(_ arguments: [String: Any], result: FlutterResult) {
    guard let seconds = arguments["seconds"] as? Int, seconds >= 1 else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setAttributionWindow: seconds should not be null or non-positive",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setAttributionWindow(seconds)
    result(nil)
  }
  
  public func setClickAttributionWindow(_ arguments: [String: Any], result: FlutterResult) {
    guard let seconds = arguments["seconds"] as? Int, seconds >= 1 else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "setClickAttributionWindow: seconds should not be null or non-positive",
        details: nil
      ))
      return
    }
    
    QGSdk.getSharedInstance().setClickAttributionWindow(seconds)
    result(nil)
  }

  public func getRecommendation(_ arguments: [String: Any], result: @escaping FlutterResult) {
    guard let scenarioId = arguments["scenarioId"] as? String, !scenarioId.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): scenarioId should not be null or empty",
        details: nil
      ))
      return
    }
    let productId = arguments["productId"] as? String
    let parameters = arguments["parameters"] as? [String: Any]
    getRecommendation(scenarioId, productId, parameters, completion: { (res) in
      if let res = res {
        do {
          try result(String(decoding: JSONSerialization.data(withJSONObject: res), as: UTF8.self))
        } catch {
          result(FlutterError(
            code: "UNEXPECTED_ERROR",
            message: "\(error)",
            details: nil
          ))
        }
      } else {
        result(nil)
      }
    })
  }

  private func getRecommendation(_ scenarioId: String, _ productId: String?, _ parameters: [String: Any]?, completion: @escaping (Any?) -> Void) {
    if let productId = productId {
      QGSdk.getSharedInstance().getRecommendationWithScenarioId(scenarioId, withProductId: productId, withQueryParameters: parameters, withCompletionHandler: completion)
    } else {
      QGSdk.getSharedInstance().getRecommendationWithScenarioId(scenarioId, withQueryParameters: parameters, withCompletionHandler: completion)
    }
  }

  public func getRecommendationPopularTags(_ arguments: [String: Any], result: @escaping FlutterResult) {
    var error: NSError? = nil
    let builder = RecommendationPopularTagsRequest.Builder()
    print("getRecommendationPopularTags")
    print(arguments)

    if let scenarioId = arguments["scenarioId"] as? String {
      builder.setScenarioId(scenarioId)
    }
    if let numItemsString = arguments["numItems"] as? String,
       let numItems = Int(numItemsString) {
      builder.setNumOfReturn(numItems)
    }

    guard let request = builder.build(error: &error) else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(String(describing: error))",
        details: nil
      ))
      return
    }

    getTagRecommendation(with: request) { res, err in
      if let err = err {
        result(err)
      } else {
        result(res)
      }
    }
  }

  public func getRecommendationProductToTags(_ arguments: [String: Any], result: @escaping FlutterResult) {
    var error: NSError? = nil
    let builder = RecommendationProductToTagsRequest.Builder()

    if let scenarioId = arguments["scenarioId"] as? String {
      builder.setScenarioId(scenarioId)
    }
    if let numItemsString = arguments["numItems"] as? String,
       let numItems = Int(numItemsString) {
      builder.setNumOfReturn(numItems)
    }
    if let pid = arguments["productId"] as? String {
      builder.setProductId(pid)
    }

    guard let request = builder.build(error: &error) else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(String(describing: error))",
        details: nil
      ))
      return
    }

    getTagRecommendation(with: request) { res, err in
      if let err = err {
        result(err)
      } else {
        result(res)
      }
    }
  }

  public func getRecommendationTagToTags(_ arguments: [String: Any], result: @escaping FlutterResult) {
    var error: NSError? = nil
    let builder = RecommendationTagToTagsRequest.Builder()

    if let scenarioId = arguments["scenarioId"] as? String {
      builder.setScenarioId(scenarioId)
    }
    if let numItemsString = arguments["numItems"] as? String,
       let numItems = Int(numItemsString) {
      builder.setNumOfReturn(numItems)
    }
    if let tagName = arguments["tagName"] as? String {
      builder.setTagName(tagName)
    }

    guard let request = builder.build(error: &error) else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(String(describing: error))",
        details: nil
      ))
      return
    }

    getTagRecommendation(with: request) { res, err in
      if let err = err {
        result(err)
      } else {
        result(res)
      }
    }
  }

  public func getRecommendationUserToTags(_ arguments: [String: Any], result: @escaping FlutterResult) {
    var error: NSError? = nil
    let builder = RecommendationUserToTagsRequest.Builder()

    if let scenarioId = arguments["scenarioId"] as? String {
      builder.setScenarioId(scenarioId)
    }
    if let numItemsString = arguments["numItems"] as? String,
       let numItems = Int(numItemsString) {
      builder.setNumOfReturn(numItems)
    }
    if let userId = arguments["userId"] as? String {
      builder.setUserId(userId)
    }

    guard let request = builder.build(error: &error) else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(String(describing: error))",
        details: nil
      ))
      return
    }

    getTagRecommendation(with: request) { res, err in
      if let err = err {
        result(err)
      } else {
        result(res)
      }
    }
  }

  public func getRecommendationTagToProducts(_ arguments: [String: Any], result: @escaping FlutterResult) {
    var error: NSError? = nil
    let builder = RecommendationTagToProductsRequest.Builder()

    if let scenarioId = arguments["scenarioId"] as? String {
      builder.setScenarioId(scenarioId)
    }
    if let numItemsString = arguments["numItems"] as? String,
       let numItems = Int(numItemsString) {
      builder.setNumOfReturn(numItems)
    }
    if let pageNumberString = arguments["pageNumber"] as? String,
       let pageNumber = Int(pageNumberString) {
      builder.setPageNum(pageNumber)
    }
    if let tags = arguments["tags"] as? String {
      builder.setTags(tags)
    }
    if let seedString = arguments["seed"] as? String,
       let seed = Double(seedString) {
      builder.setSeed(seed)
    }

    guard let request = builder.build(error: &error) else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(String(describing: error))",
        details: nil
      ))
      return
    }

    getProductRecommendation(with: request) { res, err in
      if let err = err {
        result(err)
      } else {
        result(res)
      }
    }
  }

  private func getTagRecommendation(with request: RecommendationRequest, completion: @escaping (String?, FlutterError?) -> Void) {
    QGSdk.getSharedInstance().getRecommendationTags(with: request) { response in
      if let response = response {
        do {
          let collection = try JSONSerialization.data(withJSONObject: response)
          let result = String(decoding: collection, as: UTF8.self)
          completion(result, nil)
        } catch {
          let err = FlutterError(
            code: "UNEXPECTED_ERROR",
            message: "\(error)",
            details: nil
          )
          completion(nil, err)
        }
      } else {
        completion(nil, nil)
      }
    }
  }

  private func getProductRecommendation(with request: RecommendationRequest, completion: @escaping (String?, FlutterError?) -> Void) {
    QGSdk.getSharedInstance().getRecommendationTags(with: request) { response in
      if let response = response {
        do {
          let collection = try JSONSerialization.data(withJSONObject: response)
          let result = String(decoding: collection, as: UTF8.self)
          completion(result, nil)
        } catch {
          let err = FlutterError(
            code: "UNEXPECTED_ERROR",
            message: "\(error)",
            details: nil
          )
          completion(nil, err)
        }
      } else {
        completion(nil, nil)
      }
    }
  }

  public func logRecommendationClicked(_ arguments: [String: Any], result: FlutterResult) {
    guard let scenarioId = arguments["scenarioId"] as? String, !scenarioId.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): scenarioId should not be null or empty",
        details: nil
      ))
      return
    }
    guard let modelId = arguments["modelId"] as? Int, modelId >= 1 else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): seconds should not be null or non-positive",
        details: nil
      ))
      return
    }
    guard let productId = arguments["productId"] as? String, !productId.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): productId should not be null or empty",
        details: nil
      ))
      return
    }
    guard let recommendationId = arguments["recommendationId"] as? String, !recommendationId.isEmpty else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): recommendationId should not be null or empty",
        details: nil
      ))
      return
    }
    QGSdk.getSharedInstance().logRecommendationClicked(withScenarioId: scenarioId, withModelId: modelId, withProductId: productId, withRecommendationId: recommendationId)
    result(nil)
  }

  public func hideInAppCampaigns(_ arguments: [String: Any], result: FlutterResult) {
    QGSdk.getSharedInstance().hideInAppCampaigns()
    result(nil)
  }

  public func enableStoredNotifications(_ arguments: [String: Any], result: FlutterResult) {
    QGSdk.getSharedInstance().enablePushNotificationStorage()
    result(nil)
  }

  public func getStoredNotifications(_ arguments: [String: Any], result: @escaping FlutterResult) {
    let storedNotifications = QGSdk.getSharedInstance().fetchSavedPushNotifications() ?? []
    do {
      try result(String(decoding: JSONSerialization.data(withJSONObject: storedNotifications), as: UTF8.self))
    } catch {
      result(FlutterError(
        code: "UNEXPECTED_ERROR",
        message: "\(error)",
        details: nil
      ))
    }
  }

  public func setStoredNotificationsLimit(_ arguments: [String: Any], result: @escaping FlutterResult) {
    guard let limit = arguments["limit"] as? Int else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): limit should not be null",
        details: nil
      ))
      return
    }
    QGSdk.getSharedInstance().setPushNotificationStorageLimit(limit)
    result(nil)
  }

  public func deleteAllStoredNotifications(_ arguments: [String: Any], result: @escaping FlutterResult) {
    QGSdk.getSharedInstance().deleteSavedPushNotifications()
    result(nil)
  }

  public func deleteStoredNotification(_ arguments: [String: Any], result: @escaping FlutterResult) {
    guard let index = arguments["index"] as? UInt else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): index should not be null",
        details: nil
      ))
      return
    }
    QGSdk.getSharedInstance().deleteNotification(at: index)
    result(nil)
  }

  public func getDataTrackingConfig(_ arguments: [String: Any], result: @escaping FlutterResult) {
    var jsonDictionary = [String: Any]()
    var configJsonDictionary = [String: Bool]()
    configJsonDictionary["collectLocation"] = QGSdk.getSharedInstance().dataTrackingConfig.isCollectLocation
    configJsonDictionary["collectIdfa"] = QGSdk.getSharedInstance().dataTrackingConfig.isCollectIDFA
    jsonDictionary["default"] = configJsonDictionary
    var jsonData: Data?
    do {
      jsonData = try? JSONSerialization.data(withJSONObject: jsonDictionary)
    } catch {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): JSONSerialization error: \(error)",
        details: nil
      ))
      return
    }
    guard let jsonData = jsonData else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): JSONSerialization data is nil",
        details: nil
      ))
      return
    }
    guard let jsonString = String(data: jsonData, encoding: .ascii) else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): JSON data cannot be encoded to JSON string",
        details: nil
      ))
      return
    }
    do {
      try result(jsonString)
    } catch {
      result(FlutterError(
        code: "UNEXPECTED_ERROR",
        message: "\(error)",
        details: nil
      ))
    }
  }

  public func setDataTrackingConfig(_ arguments: [String: Any], result: @escaping FlutterResult) {
    guard let config = arguments["config"] as? String else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): config should not be nil",
        details: nil
      ))
      return
    }
    guard let data = config.data(using: .utf8) else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): JSON string cannot be converted to JSON data",
        details: nil
      ))
      return
    }
    var jsonDictionary: [String: Any]?
    do {
      jsonDictionary = try JSONSerialization.jsonObject(with: data, options: .mutableContainers) as? [String: Any]
    } catch {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): JSONSerialization error: \(error)",
        details: nil
      ))
      return
    }
    guard let jsonDictionary = jsonDictionary else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): JSON dictionary is nil",
        details: nil
      ))
      return
    }
    guard let defaultDictionary = jsonDictionary["default"] as? [String: Any] else {
      result(FlutterError(
        code: "ILLEGAL_ARGUMENT",
        message: "\(#function): Default dictionary is nil",
        details: nil
      ))
      return
    }
    QGSdk.getSharedInstance().dataTrackingConfig.isCollectLocation = defaultDictionary["collectLocation"] as? Bool ?? false
    QGSdk.getSharedInstance().dataTrackingConfig.isCollectIDFA = defaultDictionary["collectIdfa"] as? Bool ?? false
    result(nil)
  }
}
